@@ -15,10 +15,23 @@
 #ifndef __DEFKEYS_H__
 #define __DEFKEYS_H__
 
+#ifdef __powerpc__
+
 #define BUTTON2MOD	0
 #define BUTTON2KEY	KEY_F10
 #define BUTTON3MOD	0
 #define BUTTON3KEY	KEY_F11
+
+#else
+
+#define BUTTON2MOD	0
+#define BUTTON2KEY	0
+#define BUTTON3MOD	0
+#define BUTTON3KEY	0
+
+#endif
+
 #define SCROLLMOD	KEY_LEFTALT
 
+
 #endif 
