@@ -18,11 +18,14 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
 #include <unistd.h>
+#include <errno.h>
 #include <fcntl.h>
 #include <string.h>
 #include <signal.h>
 #include <wait.h>
+#include <syslog.h>
 #include <sys/time.h>
 #include <linux/input.h>
 #include <linux/uinput.h>
@@ -47,12 +50,25 @@ static int ui_keyb_fd		= -1;
 
 static int running 		= -1;
 volatile sig_atomic_t		answer = 1;
+volatile sig_atomic_t		rescan = 0;
 pid_t pid		= -1;
-#define EVENT_DEVS 6
+#define EVENT_DEVS 32
 static kdev eventdevs[EVENT_DEVS];
 static input_handler ihandler[EVENT_DEVS];
 
+static int debug	= 0;
+static int autorescan   = 0;
 
+/* print debug messages to syslog or stderr */
+void debugf(const char *format, ...) {
+	va_list ap;
+	
+	if (debug) {
+		va_start(ap, format);	
+		vsyslog(LOG_DEBUG, format, ap);
+		va_end(ap);
+	}
+}
 
 static void send_event(int fd, int type, int code, int value)
 {
@@ -174,7 +190,7 @@ void keyboard_handler (int fd)
 	/* I think its best not to pass scroll, or experiment with not passing the release if
 	 * we actually used it for scrolling (but some apps may get stuck?)
 	 */
-		if (inp.code != b2_key && inp.code != b3_key && !b2_mod_pressed && !b3_mod_pressed) {
+		if (inp.code != b2_key && inp.code != b3_key) {
 			passthrough(ui_keyb_fd, inp);
 		}
 	}
@@ -220,12 +236,18 @@ void scan_for_devs()
 			ioctl(fd, EVIOCGBIT(0, EV_MAX), bit);
 			if (test_bit(EV_KEY, bit) && test_bit(EV_REP, bit)) {
 				ioctl(fd, EVIOCGID, id);
+				/* our own virtual keyboard (on rescans)*/
+				if (id[ID_PRODUCT] == 0x1F && id[ID_VENDOR] == 0x1F) {
+					close(fd);
+					continue;
+				}
 				if (id[ID_PRODUCT] != eventdevs[m].product ||
 					id[ID_VENDOR]  != eventdevs[m].vendor) {
 					if (eventdevs[m].handle >= 0) {
 						unregister_inputhandler(eventdevs[m].handle);
 						close(eventdevs[m].handle);
 					}
+					debugf("keyboard: fd %d event%d, vendor %4x product %4x\n", fd, n, id[ID_VENDOR], id[ID_PRODUCT]);
 					eventdevs[m].handle= fd;
 					eventdevs[m].product = id[ID_PRODUCT];
 					eventdevs[m].vendor = id[ID_VENDOR];
@@ -234,12 +256,18 @@ void scan_for_devs()
 				m++;
 			} else if (test_bit(EV_REL, bit)) {
 				ioctl(fd, EVIOCGID, id);
+				/* our own virtual mouse (on rescans)*/
+				if (id[ID_PRODUCT] == 0x1E && id[ID_VENDOR] == 0x1F) {
+					close(fd);
+					continue;
+				}
 				if (id[ID_PRODUCT] != eventdevs[m].product ||
 					id[ID_VENDOR]  != eventdevs[m].vendor) {
 					if (eventdevs[m].handle >= 0) {
 						unregister_inputhandler(eventdevs[m].handle);
 						close(eventdevs[m].handle);
 					}
+					debugf("mouse   : fd %d event%d, vendor %4x product %4x\n", fd, n, id[ID_VENDOR], id[ID_PRODUCT]);
 					eventdevs[m].handle= fd;
 					eventdevs[m].product = id[ID_PRODUCT];
 					eventdevs[m].vendor = id[ID_VENDOR];
@@ -257,6 +285,24 @@ void scan_for_devs()
 	}
 }
 
+void rescan_devs()
+{
+	int i, cfd;
+
+        for (i=0; i<EVENT_DEVS; i++) {
+		if (ihandler[i].fd != -1) {
+			cfd=ihandler[i].fd;
+			unregister_inputhandler(ihandler[i].fd);
+			close(cfd);
+		}
+		eventdevs[i].product = 0;
+		eventdevs[i].vendor  = 0;
+		eventdevs[i].handle  = -1;
+	}
+	usleep(100);
+	scan_for_devs();
+}
+
 int register_inputhandler (int fd, void (*func)(int fd), int grab)
 {
 	int n;
@@ -296,7 +342,7 @@ int create_fdset (fd_set *watchset)
 
 	FD_ZERO(watchset);
 	for (maxfd=n=0; n < EVENT_DEVS; n++) {
-		if (ihandler[n].fd == -1) break;
+		if (ihandler[n].fd == -1) continue;
 		FD_SET(ihandler[n].fd, watchset);
 		if (ihandler[n].fd > maxfd)
 			maxfd = ihandler[n].fd;
@@ -309,7 +355,7 @@ void call_inputhandler(fd_set *inset)
 	int n;
 
 	for (n=0; n < EVENT_DEVS; n++) {
-		if (ihandler[n].fd == -1) break;
+		if (ihandler[n].fd == -1) continue;
 		if (FD_ISSET(ihandler[n].fd, inset))
 			ihandler[n].handler (ihandler[n].fd);
 	}
@@ -333,27 +379,27 @@ int uinput_open_device(void)
 {
 	int fd = -1;
 
-	printf("Trying to open %s...", uinputdev);
+	syslog(LOG_NOTICE, "Trying to open %s...", uinputdev);
 	fd = open (uinputdev, O_RDWR);
-	printf(" %s.\n", (fd > 0)?"ok":"error");
+	syslog(LOG_NOTICE, " %s.\n", (fd > 0)?"ok":"error");
 	if (fd > 0)
 		return fd;
 
-	printf("Trying to open /dev/uinput...");
+	syslog(LOG_NOTICE, "Trying to open /dev/uinput...");
 	fd = open("/dev/uinput", O_RDWR);
-	printf(" %s.\n", (fd > 0)?"ok":"error");
+	syslog(LOG_NOTICE, " %s.\n", (fd > 0)?"ok":"error");
 	if (fd > 0)
 		return fd;
 
-	printf("Trying to open /dev/input/uinput...");
+	syslog(LOG_NOTICE, "Trying to open /dev/input/uinput...");
 	fd = open("/dev/input/uinput", O_RDWR);
-	printf(" %s.\n", (fd > 0)?"ok":"error");
+	syslog(LOG_NOTICE, " %s.\n", (fd > 0)?"ok":"error");
 	if (fd > 0)
 		return fd;
 
-	printf("Trying to open /dev/misc/uinput...");
+	syslog(LOG_NOTICE, "Trying to open /dev/misc/uinput...");
 	fd = open("/dev/misc/uinput", O_RDWR);
-	printf(" %s.\n", (fd > 0)?"ok":"error");
+	syslog(LOG_NOTICE, " %s.\n", (fd > 0)?"ok":"error");
 	if (fd > 0)
 		return fd;
 
@@ -457,17 +503,18 @@ int uinput_setup(void)
 
 void uinput_cleanup()
 {
-	int i;
+	int i, cfd;
 	
-	printf("mouseemu: cleaning...\n");
+	syslog(LOG_NOTICE, "mouseemu: cleaning...\n");
 
         uinput_close(ui_keyb_fd);
         uinput_close(ui_mouse_fd);
 
         for (i=0; i<EVENT_DEVS; i++) {
 		if (ihandler[i].fd != -1) {
+			cfd=ihandler[i].fd;
 			unregister_inputhandler(ihandler[i].fd);
-			close(ihandler[i].fd);
+			close(cfd);
 		}
 	}
 
@@ -485,6 +532,11 @@ void monitor(int sig_num)
 {
 	if (sig_num == SIGUSR1) {
 		answer = 1;
+	} else if (sig_num == SIGHUP) {
+		rescan = 1;
+	} else if (sig_num == SIGALRM) {
+		rescan = 1;
+		alarm(5);
 	} else {
 		//printf("mouseemu: aborting on sig %i \n",sig_num);	
 		/*terminate the parent:*/
@@ -508,15 +560,50 @@ void install_sighandler(void)
 
 	/*SIGUSR1 for process communication
 	 *SIGTERM and SIGCHLD for quitting
+	 *SIGHUP and SIGALRM for rescaning devices
 	 */
 	sigaction(SIGUSR1, &usr_action, NULL);
 	sigaction(SIGTERM, &usr_action, NULL);
+	sigaction(SIGHUP,  &usr_action, NULL);
 	sigaction(SIGCHLD, &usr_action, NULL);
+	sigaction(SIGALRM, &usr_action, NULL);
 	
 	sigprocmask(SIG_UNBLOCK, &mask, 0);	
 
 }
 
+/* print usage message to stdout/stderr */
+void usage(FILE *stream, char *argv[]) {
+	fprintf(stream, "usage: %s \n"
+	                "\t[-middle B2_MOD B2_KEY]\n"
+	                "\t[-right	B3_MOD B3_KEY]\n"
+	                "\t[-scroll SCROLL_MOD]\n"
+	                "\t[-typing-block DELAY]\n"
+	                "\t[-device UINPUT_DEVICE]\n"
+	                "\t[-nofork]\n"
+			"\t[-autorescan]\n"
+			"\t[-debug]\n",
+					argv[0]);
+	fprintf(stream, "All modifier and button key arguments are\n"
+	                "key scancodes. They can be found in \n"
+	                "/usr/src/linux/include/linux/input.h,\n"
+	                "or by using `showkey` in a console.\n"
+	                "Use decimal values. BTN_LEFT(272) is usable as "
+	                "B2_KEY or B3_KEY.\n\n");
+	fprintf(stream, "Default uinput device: " DEFAULT_UINPUT ".\n");
+	fprintf(stream, "Default keys:\n"
+#ifdef __powerpc__
+	                "\tMiddle click : F10 (0 68)\n"
+	                "\tRight click	: F11 (0 87)\n"
+#else
+	                "\tMiddle click : none (0 0)\n"
+	                "\tRight click	: none (0 0)\n"
+#endif
+	                "\tScroll mod.	: Alt (56)\n"
+	                "\tDefault blocking time while typing: 300ms\n");
+
+	exit(0);
+}
 
 int main(int argc, char *argv[])
 {
@@ -527,94 +614,94 @@ int main(int argc, char *argv[])
 	int nofork = 0;
 	//int argv0size = strlen(argv[0]);
 
-	printf("mouseemu " VERSION " (C) Colin Leroy <colin@colino.net>\n");
-
 	install_sighandler();
 
 	uinputdev = DEFAULT_UINPUT;
 	if (argc > 1) {
 		int i = 0;
 		if (!strcmp(argv[1],"-help")) {
-err:
-			printf("usage: %s \n"
-					"\t[-middle B2_MOD B2_KEY]\n"
-			       		"\t[-right  B3_MOD B3_KEY]\n"
-					"\t[-scroll SCROLL_MOD]\n"
-					"\t[-typing-block DELAY]\n"
-					"\t[-device UINPUT_DEVICE]\n"
-					"\t[-nofork]\n",
-			       argv[0]);
-			printf("Key codes can be found in "
-			       "/usr/src/linux/include/linux/input.h,\n"
-			       "or by using `showkey` in console.\n"
-			       "Use decimal values. BTN_LEFT(272) is usable as "
-			       "B2_KEY or B3_KEY.\n\n");
-			printf("Default uinput device: " DEFAULT_UINPUT ".\n");
-			printf("Default keys:\n"
-					"\tMiddle click : F10 (0 68)\n"
-					"\tRight click  : F11 (0 87)\n"
-					"\tScroll mod.  : Alt (56)\n"
-					"\tDefault blocking time while typing: 300ms\n");
-			
-			exit(0);
+			usage(stdout, argv);	
 		} else {
-			for (i = 1; i < argc; i++) {
-				int j = i+1;
+			i = 1;
+			while (i < argc) {
 				if (!strcmp(argv[i], "-middle")) {
-					if (argc > j+1) {
-						b2_mod = atoi(argv[j]);
-						b2_key = atoi(argv[j+1]);
+					if (argc > i+2) {
+						b2_mod = atoi(argv[i+1]);
+						b2_key = atoi(argv[i+2]);
+						i += 3;
 					} else 
-						goto err;
+						usage(stderr, argv);
 					continue;
 				} 
-				if (!strcmp(argv[i], "-right")) {
-					if (argc > j+1) {
-						b3_mod = atoi(argv[j]);
-						b3_key = atoi(argv[j+1]);
+				else if (!strcmp(argv[i], "-right")) {
+					if (argc > i+2) {
+						b3_mod = atoi(argv[i+1]);
+						b3_key = atoi(argv[i+2]);
+						i += 3;
 					} else 
-						goto err;
+						usage(stderr, argv);
 					continue;
 				}
-				if (!strcmp(argv[i], "-scroll")) {
-					if (argc > j) {
-						scroll_mod = atoi(argv[j]);
+				else if (!strcmp(argv[i], "-scroll")) {
+					if (argc > i+1) {
+						scroll_mod = atoi(argv[i+1]);
+						i += 2;
 					} else 
-						goto err;
+						usage(stderr, argv);
 					continue;					
 				}
-				if (!strcmp(argv[i], "-typing-block")) {
-					if (argc > j) {
-						typing_block_delay = atoi(argv[j]);
+				else if (!strcmp(argv[i], "-typing-block")) {
+					if (argc > i+1) {
+						typing_block_delay = atoi(argv[i+1]);
+						i += 2;
 					} else 
-						goto err;
+						usage(stderr, argv);
 					continue;					
 				}
-				if (!strcmp(argv[i], "-device")) {
-					if (argc > j) {
-						uinputdev = argv[j];
+				else if (!strcmp(argv[i], "-device")) {
+					if (argc > i+1) {
+						uinputdev = argv[i+1];
+						i += 2;
 					} else 
-						goto err;
+						usage(stderr, argv);
 					continue;					
 				}
-				if (!strcmp(argv[i], "-nofork")) {
+				else if (!strcmp(argv[i], "-nofork")) {
 					nofork=1;
+					i += 1;
 					continue;
 				}
+				else if (!strcmp(argv[i], "-autorescan")) {
+					autorescan=1;
+					i += 1;
+					continue;
+				}
+				else if (!strcmp(argv[i], "-debug")) {
+					debug=1;
+					i += 1;
+					continue;
+				} else {
+					usage(stderr, argv);
+                }
 			}
 		}
 	}
-	printf("using (%d+%d) as middle button, (%d+%d) as right button, (%d) as scroll.\n",
-		b2_mod, b2_key, b3_mod, b3_key, scroll_mod);
-	printf("using %s.\n", uinputdev);
-
 	
+	if (nofork) 
+		openlog("mouseemu", LOG_NDELAY | LOG_PERROR | LOG_PID, LOG_DAEMON); 
+	else
+		openlog("mouseemu", LOG_NDELAY | LOG_PID, LOG_DAEMON);
+	
+	syslog(LOG_NOTICE, "mouseemu " VERSION " (C) Colin Leroy <colin@colino.net>\n");
+	syslog(LOG_NOTICE, "using (%d+%d) as middle button, (%d+%d) as right button, (%d) as scroll.\n",
+		b2_mod, b2_key, b3_mod, b3_key, scroll_mod);
+
 	if (nofork)
 		goto startops;
 
 	fpid = fork();
 	if (fpid == -1) {
-		printf("can't fork\n");
+		syslog(LOG_NOTICE, "can't fork\n");
 		goto startops;
 	}
 	if (fpid != 0) {
@@ -624,7 +711,7 @@ err:
 	setsid();
 	pid = fork();
 	if (pid == -1) {
-		printf("can't fork\n");
+		syslog(LOG_NOTICE, "can't fork\n");
 		goto startops;
 	}
 
@@ -636,7 +723,13 @@ err:
 		 */
 			
 		sigset_t mask, oldmask;
+		FILE *pidfile;
 			       
+		/* write PID file so the user can signal us for device rescans */
+		pidfile = fopen("/var/run/mouseemu.pid", "w");
+		fprintf(pidfile, "%d\n", pid);
+		fclose(pidfile);
+
 		/*we start only after we received the first sigusr1 from child:*/
 
 		sigemptyset(&mask);
@@ -658,7 +751,7 @@ err:
 			}
 			
 		}
-		printf("terminating, %i\n",answer);	
+		syslog(LOG_NOTICE, "terminating, %i\n",answer);	
 		if (kill(pid, SIGTERM)<0)
 			perror("mouseemu: termination of uinput handlers failed\n");
 			
@@ -670,6 +763,7 @@ err:
 	
 	//strncpy(argv[0],"mouseemu",argv0size);
 startops:
+
 	for (i=0; i<EVENT_DEVS; i++) {
 		eventdevs[i].handle = -1;
 		eventdevs[i].vendor = 0;
@@ -685,28 +779,45 @@ startops:
 
 	running = uinput_setup();
 	if (running < 0) {
-		printf("mouseemu: Make sure uinput module is loaded or available "
-			"in the kernel.\n");
+		syslog(LOG_NOTICE, "Make sure uinput module is loaded or available "
+		                   "in the kernel.\n");
 	}
                                                          
 
 	chdir("/");
-
+	
+	if (autorescan) 
+		alarm(5);
+	
 	/*main loop*/
 	
         while(running > 0) {
 	
 		tv.tv_sec = 1; tv.tv_usec = 0;
 		maxfd = create_fdset(&inset);
-		if ((val = select (maxfd+1, &inset, NULL, NULL, &tv)) >= 0) {
+		val = select (maxfd+1, &inset, NULL, NULL, &tv);
+		/* signal received, so rescan for devices when idle*/
+		if (val == 0 && rescan) {
+			rescan = 0;
+			rescan_devs();
+		}
+		if (val >= 0) {
 			if (val == 0)
 				usleep(10);
-			else
-				call_inputhandler(&inset);
+			else {
+				if (errno == ENODEV) {
+					debugf("device disconnect detected (select %d, errno %d), rescanning devices\n", val, errno);
+					errno = 0;
+					rescan_devs();
+					usleep(500);
+				} else {
+					call_inputhandler(&inset);
+				}
+			}
 		}
 		/* tell the parent we are running without problems */
 		/* What should we do if the parent is dead? */
-		if (answer) {
+		if (answer && !nofork) {
 			answer=0;
 			kill(getppid(), SIGUSR1);
 		}
